import { Point } from "../../common/point";
import { getRoot } from "pi_common_tools/ui/main_root";
import { EventTypes } from "pi_gui/engine/event_base";
import { getDocument } from "pi_gui/main";
import { MapRenderScene } from "../render/map_scene";
import { AdaptationManage } from "./adaptation";
import { IEventData, IMouseCallbacks, IScene3DEventInfo } from "../../interface/screen_con";
import { registerMultiPointerEvent, unregisterMultiPointerEvent } from "pi_load/extensions/multi_pointers";

class MouseEventCallBack {
    constructor() {
        return MouseEventCallBack.createdProxy(this);
    }

    private static __funcNameMap: WeakMap<Function, string> = new WeakMap();

    private static createdProxy = (example) => {
        return new Proxy(example, {
            get(target, method: string, receiver) {
                let executor = (...args) => {
                    let list = Reflect.get(target, method, receiver) || [];
                    if (!list.length) return;
                    list.forEach(cb => cb(...args));
                }
                if (method === 'remove') { //特殊函数 移除回调
                    executor = (cb) => {
                        let funcName = MouseEventCallBack.__funcNameMap.get(cb);
                        let list = Reflect.get(target, funcName, receiver) || [];
                        if (!list.includes(cb)) return false;
                        list.splice(list.indexOf(cb), 1);
                        return true;
                    }
                }
                else if (method === 'clear') {
                    executor = () => {
                        Object.keys(target).forEach(key => {
                            Reflect.deleteProperty(target, key);
                        });
                        console.log('clear callbacks--->', target);
                        return true
                    }
                }
                return executor
            },
            set(target, method: string, cb, receiver) {
                let list = Reflect.get(target, method, receiver) || [];
                if (list.includes(cb)) return false
                list.push(cb);
                MouseEventCallBack.__funcNameMap.set(cb, method);
                return Reflect.set(target, method, list, receiver);
            }
        }) as IMouseCallbacks
    }
}

export class MapMouseControl {
    //#region 绑定事件
    private readonly eventCallBack: IMouseCallbacks = new MouseEventCallBack();

    /**
     * 绑定事件
     */
    public bindMouseCallBack = (options: IMouseCallbacks) => {
        if (!options) return;
        for (let key in options) {
            this.eventCallBack[key] = options[key];
        }
    }

    /**
     * 解绑事件
     */
    public unBindCallBack = (options: IMouseCallbacks) => {
        if (!options) return;
        for (let key in options) {
            this.eventCallBack["remove"](options[key]);
        }
    }
    //#endregion

    //#region 是否启用鼠标滚轮
    private isMouseWheelActive: boolean = true;

    public activeMouseWheel(flag: boolean) {
        this.isMouseWheelActive = flag;
    }
    //#endregion

    private leftPointer: { x: number, y: number } = undefined; // 左键按下记录
    private rightPointer: { x: number, y: number } = undefined; // 右键按下记录
    private realMove: boolean = false; // 是否真实移动过
    private enableHover: boolean = true; // 如果按下的话，不启用滑动
    private pixelRatio: number;

    constructor(private mapRenderScene?: MapRenderScene) {
        this.active();
        this.pixelRatio = AdaptationManage.getScreenInfo().pixelRatio;
    }

    /**
     * 鼠标离开事件
     */
    private onMouseLeave = (event: IEventData) => {
        this.eventCallBack.onLeave();
        // this.onPoiterUp(event);
    }

    /**
     * 将 GUI 事件数据格式化成项目自定义的
     * @example 加上 prx pry
     */
    private eventDataFormat = (event: IEventData): IEventData => {
        let prx = event.pageX * this.pixelRatio;
        let pry = event.pageY * this.pixelRatio;
        return { ...event, prx, pry };
    }

    /**
     * ui 鼠标按下移动
     */
    private onPointerDownMove = (ev: IEventData) => {
        ev = this.eventDataFormat(ev);
        let movex: number = 0, movey: number = 0;
        if (this.leftPointer) {
            movex = (ev.pageX - this.leftPointer.x) * this.pixelRatio;
            movey = (ev.pageY - this.leftPointer.y) * this.pixelRatio;
            this.leftPointer = { x: ev.pageX, y: ev.pageY };
        }
        else if (this.rightPointer) {
            movex = (ev.pageX - this.rightPointer.x);
            movey = (ev.pageY - this.rightPointer.y);
            this.rightPointer = { x: ev.pageX, y: ev.pageY };
        }
        this.eventCallBack.onDownMove(ev, movex, movey);

        if (!this.mapRenderScene) return;
        this.mapRenderScene.scene.onPointerMove((<any>ev));
    }

    // /**
    //  * 鼠标不按下移动
    //  */
    // private pointerHover = (ev: IEventData) => {
    //     ev = this.eventDataFormat(ev);
    //     this.eventCallBack.onHover(ev);
    // }

    /**
     * ui 鼠标按下
     */
    private onPointerDown = (ev: IEventData) => {
        ev = this.eventDataFormat(ev);
        if (ev.button === 0 && !this.rightPointer) {
            this.leftPointer = { x: ev.pageX, y: ev.pageY };
        }
        else if (ev.button === 2 && !this.leftPointer) {
            this.rightPointer = { x: ev.pageX, y: ev.pageY };
        }
        this.eventCallBack.onDown(ev);

        if (!this.mapRenderScene) return;
        this.mapRenderScene.scene.onPointerDown((<any>ev));
    }

    /**
     * ui 鼠标抬起
     */
    private onPoiterUp = (ev: IEventData) => {
        ev = this.eventDataFormat(ev);
        this.leftPointer = undefined;
        this.rightPointer = undefined;
        this.realMove = false;
        this.eventCallBack.onUp(ev);

        if (!this.mapRenderScene) return;
        this.mapRenderScene.scene.onPointerUp((<any>ev));
    }

    /**
     * ui 鼠标点击
     */
    private onPointerClick = (ev: IEventData) => {
        ev = this.eventDataFormat(ev);
        this.eventCallBack.onClick(ev);

        if (!this.mapRenderScene) return;
        this.mapRenderScene.scene.onPointerClick((<any>ev));
    }

    /**
     * 滚轮滑动
     * @param ev 
     * @returns 
     */
    private onMouseWheel = (ev: IEventData) => {
        //#region 这里直接把 hexLayout 中的目标缩放计算好
        let direction: "up" | "down" = ev.deltaY > 0 ? "up" : "down";
        // let scaleValue = direction === "up" ? -0.01 : 0.01;
        // let targetScale: number = this.hexLayout.scaleZoom(scaleValue);
        //#endregion
        let targetScale: number = 1;
        this.eventCallBack.onWheel(direction, targetScale);
        if (!this.isMouseWheelActive) return;
    }

    /**
     * 场景事件按下点
     */
    private sceneDownPoint: Point;

    /**
     * 场景点击
     */
    private onPointerClickScene = (e: IScene3DEventInfo): void => {
        let pickInfo = this.mapRenderScene.scenePlanePick(e);
        let pickPoint = new Point(pickInfo.x, pickInfo.y);
        // console.log(e.isSendNextLayer, "onPointerClickScene pick hex source", pickHex, "hex round", pickHex.round());
        if (!e.isSendNextLayer) return;
        this.eventCallBack.onSceneClick(e, pickPoint);
    }

    /**
     * 场景按下滑动
     */
    private onPointerDownMoveScene = (e: IScene3DEventInfo): void => {
        if (!this.sceneDownPoint) return;
        let pickInfo = this.mapRenderScene.scenePlanePick(e);
        let pickPoint = new Point(pickInfo.x, pickInfo.y);
        let movex = (this.sceneDownPoint.x - pickPoint.x);
        let movey = (this.sceneDownPoint.y - pickPoint.y);
        this.eventCallBack.onSceneDownMove(e, pickPoint, movex, movey);
    }

    /**
     * 场景按下
     */
    private onPointerDownScene = (e: IScene3DEventInfo): void => {
        let pickInfo = this.mapRenderScene.scenePlanePick(e);
        let pickPoint = new Point(pickInfo.x, pickInfo.y);
        this.sceneDownPoint = pickPoint;
        this.eventCallBack.onSceneMouseDown(e, pickPoint);
    }

    /**
     * 场景抬起
     */
    private onPointerUpScene = (e: IScene3DEventInfo): void => {
        if (!this.sceneDownPoint) return;
        let pickInfo = this.mapRenderScene.scenePlanePick(e);
        let pickPoint = new Point(pickInfo.x, pickInfo.y);
        this.sceneDownPoint = null;
        this.eventCallBack.onSceneMouseUp(e, pickPoint);
    }

    /**
     * 双指事件
     */
    private multiPointer = (uiDistance: number, _deltaAngle: number, ev: IEventData) => {
        if (!ev.isSendNextLayer) return;
        this.eventCallBack.onMultiPointer(uiDistance);
    }

    //#region 鼠标控制器是否激活
    public get isActive() {
        return this._isActive;
    }
    private _isActive: boolean = false;

    public active() {
        if (this.isActive) return;
        this._isActive = true;
        this.attachControl();
        console.log("MouseControl is active");
    }

    public pause() {
        if (!this.isActive) return;
        this._isActive = false;
        this.detachControl();
        console.log("MouseControl is pause");
    }

    private attachControl = () => {
        getDocument().addEventListener(EventTypes.up, this.onPoiterUp);          // 抬起
        getDocument().addEventListener(EventTypes.move, this.onPointerDownMove);   // 按下移动
        getDocument().addEventListener(EventTypes.click, this.onPointerClick);   // 点击
        getDocument().addEventListener(EventTypes.down, this.onPointerDown);     // 按下
        getDocument().addEventListener(EventTypes.wheel, this.onMouseWheel);     // 滚轮
        // getDocument().addEventListener(EventTypes.mousemove, this.pointerHover); // hover
        getDocument().addEventListener("pointerleave", this.onMouseLeave);       // 鼠标离开屏幕外（pc web平台、取消网格线高亮）
        registerMultiPointerEvent(this.multiPointer, getRoot() as any);

        if (!this.mapRenderScene) return;
        this.mapRenderScene.scene.addClickListen(this.onPointerClickScene as any);    // 场景点击
        this.mapRenderScene.scene.addMoveListen(this.onPointerDownMoveScene as any);  // 场景按下移动
        this.mapRenderScene.scene.addDownListen(this.onPointerDownScene as any);      // 场景按下
        this.mapRenderScene.scene.addUpListen(this.onPointerUpScene as any);          // 场景抬起
    }

    private detachControl = () => {
        getDocument().removeEventListener(EventTypes.up, this.onPoiterUp);          // 抬起
        getDocument().removeEventListener(EventTypes.move, this.onPointerDownMove);   // 按下移动
        getDocument().removeEventListener(EventTypes.click, this.onPointerClick);   // 点击
        getDocument().removeEventListener(EventTypes.down, this.onPointerDown);     // 按下
        getDocument().removeEventListener(EventTypes.wheel, this.onMouseWheel);     // 滚轮
        // getDocument().removeEventListener(EventTypes.mousemove, this.pointerHover);    // hover
        getDocument().removeEventListener("pointerleave", this.onMouseLeave);       // 鼠标离开屏幕外（pc web平台、取消网格线高亮）
        unregisterMultiPointerEvent();

        if (!this.mapRenderScene) return;
        this.mapRenderScene.scene.removeClickListen(this.onPointerClickScene as any);    // 场景点击
        this.mapRenderScene.scene.removeMoveListen(this.onPointerDownMoveScene as any);  // 场景按下移动
        this.mapRenderScene.scene.removeDownListen(this.onPointerDownScene as any);      // 场景按下
        this.mapRenderScene.scene.removeUpListen(this.onPointerUpScene as any);          // 场景抬起
    }
    //#endregion
}
