import { Point } from "../../common/point";
import { UpdateCallBack } from "../logic/render_call_back";
import { MapRenderScene } from "./map_scene";
import { MapMouseControl } from "../logic/map_mouse_control";
import { IEventData, IScene3DEventInfo } from "../../interface/screen_con";

export class CityMapManager {
    /**
     * 双指事件回调, 某些ui渲染事件里面，需要节流
     */
    private readonly multiPointerEvnetCallBack: UpdateCallBack<boolean> = new UpdateCallBack();
    /**
     * 双指事件结束事件 （触发了双指事件之后，一段时间内不再触发才算结束，也不往外部推送事件）
     */
    private multiPointerEvnetTime: number;
    /**
     * 双指事件状态
     */
    private multiPointerEvnetState: boolean = false;
    /**
     * 是否允许拖动
     */
    private isMove: boolean = false;
    /**
    * 比例尺缩放范围
    */
    public scaleScope: [number, number] = [500, 512];
    /**
    * 场景大小
    */
    public sceneInfo:{ width:number, height:number };
    /**
     * hex map 移动范围 [minx, maxx]
    */
    public cameraMoveScopeX: [number, number] = [-10, 10];
    /**
     * hex map 移动范围 [miny, maxy]
     */
    public cameraMoveScopeY: [number, number] = [-10, 10];
    /**
     * 当前地图缩放值, 同时也是正交相机的大小
     */
    public scale: number = 1;
    /**
     * move Record 移动记录，拖动之后 松手时 需要记录的位置，记录了当前地图容器距离屏幕左上角的距离，用来和 点击事件位置 一起计算命中
     */
    public moveRecord: Point = new Point(0, 0);
    constructor(
        public readonly mapMouseControl: MapMouseControl,
        public readonly mapRenderScene: MapRenderScene,
        scale: number,
    ) {
        this.scale = scale;
        this.initEvent();
    }


    /**
     * 设置相机信息
     */
    public setCamera(scale:number, scaleScope:[number, number], width:number, height:number) {
        this.scaleScope = scaleScope;
        this.sceneInfo = {
            width,
            height
        };
        this.scaleTo(scale);
    }
    /**
     * 移动到指定坐标
     */
    public moveTo(x: number, y: number) {
        this.mapRenderScene.moveTo(x, y);
    }

    /**
     * 绑定回调函数
     * @example 将函数包装成整个对象传进来
     */
    public bindMultiPointerCallBack(multiPointerCallBack: (multiPointerState: boolean) => void): void;
    /**
     * 解绑回调函数
     */
    public bindMultiPointerCallBack(multiPointerCallBack: (multiPointerState: boolean) => void, unBind: true): void;
    public bindMultiPointerCallBack(multiPointerCallBack: (multiPointerState: boolean) => void, unBind?: true): void {
        if (unBind) {
            multiPointerCallBack && this.multiPointerEvnetCallBack.remove(multiPointerCallBack);
        }
        else {
            multiPointerCallBack && (this.multiPointerEvnetCallBack.update = multiPointerCallBack);
        }
    }

    /**
     * 初始化注册事件
     */
    private initEvent() {
        this.mapMouseControl.bindMouseCallBack({
            onUp: this.onMouseUp,
            onDown: this.onMouseDown,
            onSceneDownMove: this.onSceneMouseDownMove,
            onWheel: this.onMouseWheel,
            onMultiPointer: this.onMultiPointer
        })

        this.moveTo(0, 0);
    }

    /**
     * 缩放到指定值
     */
    public scaleTo(scale: number) {
        this.scale = Math.max(this.scaleScope[0], Math.min(this.scaleScope[1], scale));
        this.mapRenderScene.scaleTo(this.scale);
        this.moveZoom(0, 0);
        return this.scale;
    }

    /**
     * 累计缩放
     */
    public scaleZoom(v: number) {
        this.scale += v * 40;
        return this.scaleTo(this.scale);
    }

    /**
     * 累加移动， 并改摄像机位置
     */
    public moveZoom(x: number, y: number) {
        this.moveToPosition(this.moveRecord.x + x, this.moveRecord.y + y)
    }
    /**
     * 累加移动， 并改摄像机位置
     */
    public moveToPosition(_x: number, _y: number) {
        const {x, y, change } = this.getMoveEndPosition(_x, _y);
        //位置改变才更新
        if(change){
            this.moveRecord.x = x;
            this.moveRecord.y = y;
            this.mapRenderScene.moveTo(this.moveRecord.x, this.moveRecord.y);
        }
    }
    /**
     * 获取相机移动终点
     */
    public getMoveEndPosition(x: number, y: number) {
        x = Math.min(this.cameraMoveScopeX[1], Math.max(this.cameraMoveScopeX[0], x));
        y = Math.min(this.cameraMoveScopeY[1], Math.max(this.cameraMoveScopeY[0], y));
        const change:boolean = this.moveRecord.x !== x || this.moveRecord.y !== y;
        return {x, y, change }
    }

    /**
     * 滚轮缩放
     */
    private onMouseWheel = (direction: "up" | "down") => {
        // if (!GuideKeyboardControl.isKeyBoardRespond) return;
        this.scaleZoom(direction === "up" ? 2 : -2);
    }

    private multiPointerCheck() {
        if (!this.multiPointerEvnetState || !this.multiPointerEvnetTime) return; // 双指事件没触发
        // if (Date.now() < this.multiPointerEvnetTime) return; // 距离上次触发时间不足300毫秒，不通知事件
        this.multiPointerEvnetTime = 0;
        this.multiPointerEvnetState = false;
        this.multiPointerEvnetCallBack.update(this.multiPointerEvnetState);
    }

    /**
     * 双指事件
     */
    private onMultiPointer = (deltaDistance: number) => {
        // if (CityMode.currentMode !== ECityMapMode.Default) return;
        // if (!GuideKeyboardControl.isKeyBoardRespond) return;
        deltaDistance = deltaDistance/10; 

        // console.log("=================>",deltaDistance)
        if (Math.abs(deltaDistance) > 5 || Math.abs(deltaDistance) < 0.01) return;
        // console.error("==============>deltaDistance111",deltaDistance)
        this.scaleZoom(-deltaDistance);
        if (this.multiPointerEvnetState) return;
        this.multiPointerEvnetTime = Date.now() + 1000; // 300毫秒之后自动通知结束
        this.multiPointerEvnetState = true;
        this.multiPointerEvnetCallBack.update(this.multiPointerEvnetState);
    }

    /**
     * 按下
     */
    private onMouseDown = (_event: IEventData) => {
        // if (!GuideKeyboardControl.isKeyBoardRespond) return;
        // if (!CityMode.checkMapDown()) return;
        this.isMove = true;
    }

    /**
     * 抬起
     */
    private onMouseUp = (_event: IEventData) => {
        if (!this.isMove) return;
        this.isMove = false;
        this.multiPointerCheck();
    }

    /**
     * 按下滑动事件,改成场景触发,更加准确
     */
    private onSceneMouseDownMove = (_event: IScene3DEventInfo, _pickPoint: Point, movex: number, movey: number) => {
        // if (!CityMode.checkMapMove()) return;
        if (!this.isMove) return;
        this.moveZoom(movex, movey);
    }
    /**
     * 镜头移动范围
     * @param dx 移动x
     * @param dy 移动y
     * @param camera 相机
     * @param backgroundMesh 背景图
    */
    public calcCameraRange(){
        let {width, height} = this.sceneInfo;
        const camera =  this.mapRenderScene.camera;
        const  rw = width, //面片在scene中经过比例尺变化后的width
            rh = height,//面片在scene中经过比例尺变化后的height
            rx = -Math.abs(rw/2),//左边缘
            ry = -Math.abs(rh/2);//上边缘
        let cw = (camera.orthoRight - camera.orthoLeft) * 0.5;
        let ch = (camera.orthoTop - camera.orthoBottom) * 0.5;
        this.cameraMoveScopeX = [Math.min(rx + cw, rx + rw / 2), Math.max(rx + rw - cw, rx + rw / 2)];
        this.cameraMoveScopeY = [Math.min(ry + ch, ry + rh / 2), Math.max(ry + rh - ch, ry + rh / 2)];
    }
}

