import { Point } from "../../common/point";
import { BABYLON } from "pi_babylon/render3d/babylon";
import { SceneStruct } from "pi_babylon/scene_base";
import { IScene3DEventInfo } from "../../interface/screen_con";
import { formatEventDataToScene, rayIntersectsPlane } from "./scene_utils";
import { IPointerEvent } from "../../interface/scene_con";
import { CityMap } from "../map_index";
import { BABYLONAction, CustomScene } from "pi_babylon/index";

export class MapRenderScene {
    /**
     * 激活状态
     */
    private isActive: boolean = false;
    /**
     * 相机
     */
    public readonly camera: BABYLON.FreeCamera;
    /**
     * 场景
     */
    public readonly scene: CustomScene;
    /**
     * 命中检查用的平面
     */
    private readonly plane = new BABYLON.Plane(0, 0, 1, 0);
    /**
     * 背景面片
     */
    public backgroundMesh: BABYLON.Mesh;

    /**
     * 缩放回调
     */
    public scaleCallback: Function;

    constructor(scale: number, scene: SceneStruct) {
        this.scene = scene;
        this.camera = <BABYLON.FreeCamera>this.scene.camera;
        this.camera.position = new BABYLON.Vector3(0, 0, -10);
        this.initScene();
        this.initCamera(scale);
        this.initSceneEvent();
        this.active();
    }
    //设置相机尺寸改变回调
    public setCameraScaleCallback(callback: Function) {
        this.scaleCallback = callback;
    }
    /**
     * 初始化场景
     */
    private initScene() {
        if (!this.scene.impl.metadata) {
            this.scene.impl.metadata = {};
        }
        this.scene.impl.blockMaterialDirtyMechanism = true;
        this.scene.impl.metadata.gltfAsLeftHandedSystem = true;
        this.scene.impl.autoClear = false;
        this.scene.impl.autoClearDepthAndStencil = false;
        this.scene.impl.ambientColor = BABYLON.Color3.White();
    }

    /**
     * 初始化 事件 转换
     */
    private initSceneEvent() {
        this.scene.pickCall = () => { };

        this.scene.onPointerClick = (e: IPointerEvent) => {
            let pickInfo = formatEventDataToScene(this.scene, e);
            this.scene.pointerClickListenMap.forEach(func => {
                func(pickInfo as any);
            });
        };

        this.scene.onPointerMove = (e: IPointerEvent) => {
            let pickInfo = formatEventDataToScene(this.scene, e);
            this.scene.pointerMoveListenMap.forEach(func => {
                func(pickInfo as any);
            });
        };

        this.scene.onPointerDown = (e: IPointerEvent) => {
            let pickInfo = formatEventDataToScene(this.scene, e);
            this.scene.pointerDownListenMap.forEach(func => {
                func(pickInfo as any);
            });
        };

        this.scene.onPointerUp = (e: IPointerEvent) => {
            let pickInfo = formatEventDataToScene(this.scene, e);
            this.scene.pointerUpListenMap.forEach(func => {
                func(pickInfo as any);
            });
        };
    }

    /**
     * 初始化相机
     */
    private initCamera(scale: number) {
        this.camera.setTarget(BABYLON.Vector3.Forward());
        this.camera.mode = BABYLON.FreeCamera.ORTHOGRAPHIC_CAMERA;
        this.camera.piSizeForWidth = true;
        this.camera.piOrthSize = scale;
        this.camera.minZ = 0.1;
        this.camera.maxZ = 300;
        this.camera.rotation.copyFromFloats(0, 0, 0);
    }
    public pause = () => {
        if (!this.isActive) return;
        this.isActive = false;
        this.scene.pause();
    }

    public active = () => {
        if (this.isActive) return;
        this.scene.addCamera(this.camera);
        this.scene.setCurrCamera(this.camera.name);
        this.scene.active();
        this.isActive = true;
    }

    /**
     * 视野移动到
     */
    public moveTo(moveTox: number, moveToy: number) {
        this.applyCamera(moveTox, moveToy);
    }

    public scaleTo(scale: number) {
        BABYLONAction.updateEngineDimensions(this.scene.renderWidth,this.scene.renderHeight);
        this.camera.piOrthSize = scale;
        this.scaleCallback && this.scaleCallback(scale);
        CityMap.getCityManager()?.calcCameraRange();
    }

    /**
     * 应用相机位置
     */
    private applyCamera(x: number, y: number) {
        this.camera.position.copyFromFloats(x, y, this.camera.position.z);
    }
    /**
     * 计算场景命中位置
     */
    public scenePlanePick(ev: IScene3DEventInfo): Point {
        BABYLONAction.updateEngineDimensions(this.scene.renderWidth,this.scene.renderHeight);
        let event = ev.sceneEvent;
        return this.pointToPlanePick(event.x, event.y);
    }

    /**
     * 命中点转换成场景点
     */
    public pointToPlanePick(x: number, y: number): Point {
        const hitPoint = rayIntersectsPlane(this.camera, this.scene, x, y, this.plane);
        if (!hitPoint) return new Point(0, 0);
        return new Point(hitPoint.x, hitPoint.y);
    }
}
