import { CustomGeometry, CustomScene, Constants, MainOpacityMaterial, RenderTargetFactory, BABYLONAction } from "pi_babylon/index";
import { BABYLON } from "pi_babylon/render3d/babylon";
import { CustomRenderTargetScene } from "pi_babylon/fullscreen_scene";
import { DualBlur } from "pi_babylon/postprocess/dual_blur"
import { KeyDualBlur } from "pi_babylon/interface/index";
import { RibbonBuilder } from "pi_babylon/ribbon/ribbon";
import { AreaMaterial } from "./area_material";

type TVerticeID = number;
type TPoint = [number, number];

interface IPolygon {
    triangles: [TVerticeID, TVerticeID, TVerticeID][];
    min: TPoint;
    max: TPoint;
    points: TVerticeID[];
    [index: string]: any;
}
interface IVertices {
    get(id: TVerticeID): TPoint;
}

type TLink = [TVerticeID, number, number, number];

class Links {
    protected pointLinks: Map<TVerticeID, TLink[]> = new Map();
    constructor(edges: [TVerticeID, TVerticeID][], data: IVertices) {
        let edgeCount = edges.length;
        for (let i = 0; i < edgeCount; i++) {
            let edge    = edges[i];
            let from    = edge[0];
            let to      = edge[1];
            let a       = data.get(from);
            let b       = data.get(to);
            let x = b[0] - a[0];
            let y = b[1] - a[1];
            let l = Math.sqrt(x * x + y * y);
            {
                let links = this.pointLinks.get(from);
                if (links == undefined) {
                    this.pointLinks.set(from, []);
                    links = this.pointLinks.get(from);
                }
                if (links) {
                    if (links.findIndex(item => item[0] == to) < 0) {
                        links.push([to, x / l, y / l, 0]);
                    }
                }
            }
            {
                let links = this.pointLinks.get(to);
                if (links == undefined) {
                    this.pointLinks.set(to, []);
                    links = this.pointLinks.get(to);
                }
                if (links) {
                    if (links.findIndex(item => item[0] == from) < 0) {
                        links.push([from, -x / l, -y / l, 0]);
                    }
                }
            }
        }
        this.pointLinks.forEach((links) => {
            // 按 与 x 轴逆时针夹角递增排序
            links.sort((a, b) => {
                if (a[2] >= 0 && b[2] >= 0) {
                    return a[1] - b[1] > 0 ? -1 : 1;
                } else if (a[2] < 0 && b[2] < 0){
                    return a[1] - b[1] > 0 ? 1 : -1;
                } else {
                    return a[2] - b[2] > 0 ? -1 : 1;
                }
            });
            let len = links.length;
            // 计算每边与排序后下一条边的逆时针夹角
            for (let i = 0; i < len; i++) {
                let cur  = links[i];
                let next = links[ (i + len + 1) % len ];
                let x1 =  cur[1]; let y1 =  cur[2];
                let x2 = next[1]; let y2 = next[2];
                let angle = Math.acos(x1 * x2 + y1 * y2);
                if ((x1 * y2 - x2 * y1) < 0) {
                    angle = Math.PI * 2 - angle;
                }
                cur[3] = angle;
            }
        });
    }
    public polygons(data: IVertices) {
        let pointCount = this.pointLinks.size;
        let points: TVerticeID[] = [];
        this.pointLinks.forEach((v, k) => {
            points.push(k);
        });

        // 已被确定为多边形的边的信息 - 每个有向边信息只会被一个多边形使用一次
        let findedLink: [TVerticeID, TVerticeID][] = [];
        // 一次查找经过的边信息
        let findingLink: [TVerticeID, TVerticeID][] = [];
        let polygons: TVerticeID[][] = [];
        let start: TVerticeID[] = [];
        // 以每个点为起点寻找逆时针环绕的边构成的多边形
        for (let i = 0; i < pointCount; i++) {
            let point = points[i];
            let links = this.pointLinks.get(point);
            if (links) {
                let len = links.length;
                // 分别以该点关联的边为起始边寻找
                for (let i = 0; i < len; i++) {
                    let link = links[i];
                    let next: TLink = link;
                    let cur = point;
                    let polygon: TVerticeID[] = [cur];
                    let max = 512;
                    let angle = 0;
                    // 从零记录查找经过的边信息
                    findingLink.length = 0;
                    let angleHurge = cur;
                    while (max >= 0 && next != undefined && next[0] != point) {
                        // 检查目标边是否已被多边形使用, 是则说明此次查找的多边形已记录,跳过此次查找
                        if (findedLink.findIndex((val) => { return val[0] == cur && val[1] == next[0] }) >= 0) {
                            cur = undefined;
                            break;
                        }
                        polygon.push(next[0]);
                        // console.error(cur, next[0]);
                        findingLink.push([cur, next[0]]);

                        let temp = this.findLinkNext(next, cur);
                        if (temp) {
                            cur = next[0];
                            next = temp[0];
                            angle += temp[1];
                            if (temp[1] > Math.PI) {
                                angleHurge = cur;
                            }
                        } else {
                            cur = undefined;
                            break;
                        }
                        max --;
                    }
                    
                    // console.error(cur, next[0]);
                    findingLink.push([cur, next[0]]);

                    // 当前起始边与结束边闭合, 且累计的预期内角和满足多边形内角和定理
                    if (cur != undefined && links[(i + len + 1) % len][0] == cur && angle <= (polygon.length - 2) * Math.PI) {
                        let len = polygon.length;
                        let min = polygon[0];
                        let idx = 0;
                        for (let i = 0; i < len; i++) {
                            if (min > polygon[i]) {
                                idx = i;
                                min = polygon[i];
                            }
                        }
                        let temp = [];
                        for (let i = idx; i < idx + len; i++) {
                            temp.push(polygon[i % len]);
                        }
                        findedLink.push(...findingLink);
                        // let has = polygons.findIndex((v) => { let flag = true; v.forEach((v, i) => { flag = flag && v == temp[i] }); return flag; }) >= 0;
                        // if (!has) {
                            polygons.push(temp);
                            start.push(angleHurge);
                        // }
                    }
                }
            }
        }
        // console.error(findedLink);
    
        let polygonsnew: IPolygon[] = [];
        polygons.forEach((polygon, idx) => {
            let len = polygon.length;
            let polygonnew: IPolygon = {
                triangles: [],
                points: [],
                min: [Number.MAX_SAFE_INTEGER, Number.MAX_SAFE_INTEGER],
                max: [Number.MIN_SAFE_INTEGER, Number.MIN_SAFE_INTEGER],
            }
            let idxstart = polygon.indexOf(start[idx]);
            for (let i = 0; i < len - 2; i++) {
                let a = polygon[idxstart];
                let b = polygon[(idxstart + i + 1) % len];
                let c = polygon[(idxstart + i + 2) % len];
                polygonnew.triangles.push([a, b, c]);
                let ap = data.get(a);
                let bp = data.get(b);
                let cp = data.get(c);
                polygonnew.min[0] = Math.min(ap[0], bp[0], cp[0], polygonnew.min[0]);
                polygonnew.min[1] = Math.min(ap[1], bp[1], cp[1], polygonnew.min[1]);
                polygonnew.max[0] = Math.max(ap[0], bp[0], cp[0], polygonnew.max[0]);
                polygonnew.max[1] = Math.max(ap[1], bp[1], cp[1], polygonnew.max[1]);
            }
            polygonnew.points = polygon;
            polygonsnew.push(polygonnew);
        });
        return polygonsnew;
    }
    protected findLinkNext(cur: TLink, linking: TVerticeID): [TLink, number] | undefined {
        let links = this.pointLinks.get(cur[0]);
        if (links) {
            let idx = links.findIndex((a) => a[0] == linking);
            let len = links.length;
            if (idx >= 0) {
                let link = links[(idx + len - 1) % len];
                return [link, link[3]];
            } else {
                return undefined;
            }
        }
        return undefined;
    }
    public findEdge(data: IVertices, includePoints: TVerticeID[]) {
        let pointCount = includePoints.length;
        let points: TVerticeID[] = includePoints;
    
        let pointLinks = new Map();
        includePoints.forEach((key) => {
            let links = this.pointLinks.get(key);
            let newlinks = [];
            pointLinks.set(key, newlinks);
            links.forEach((link) => {
                if (includePoints.includes(link[0])) {
                    newlinks.push(link);
                }
            });
        });

        // 已被确定为多边形的边的信息 - 每个有向边信息只会被一个多边形使用一次
        let findedLink: [TVerticeID, TVerticeID][] = [];
        // 一次查找经过的边信息
        let findingLink: [TVerticeID, TVerticeID][] = [];
        let polygons: TVerticeID[][] = [];
        let start: TVerticeID[] = [];
        // 以每个点为起点寻找逆时针环绕的边构成的多边形
        for (let i = 0; i < pointCount; i++) {
            let point = points[i];
            let links = pointLinks.get(point);

            if (polygons.length == 1) break;

            if (links) {
                let len = links.length;
                // 分别以该点关联的边为起始边寻找
                for (let i = 0; i < len; i++) {

                    if (polygons.length == 1) break;

                    let link = links[i];
                    let next: TLink = link;
                    let cur = point;
                    let polygon: TVerticeID[] = [cur];
                    let max = 512;
                    let angle = 0;
                    // 从零记录查找经过的边信息
                    findingLink.length = 0;
                    let angleHurge = cur;
                    while (max >= 0 && next != undefined && next[0] != point) {
                        // 检查目标边是否已被多边形使用, 是则说明此次查找的多边形已记录,跳过此次查找
                        if (findedLink.findIndex((val) => { return val[0] == cur && val[1] == next[0] }) >= 0) {
                            cur = undefined; break;
                        }
                        polygon.push(next[0]);
                        findingLink.push([cur, next[0]]);

                        let temp = this.findLinkNextClockwise(cur, pointLinks.get(next[0]));
                        if (temp) {
                            cur = next[0];
                            next = temp[0];
                            angle += temp[1];
                            if (temp[1] > Math.PI) {
                                angleHurge = cur;
                            }
                        } else {
                            cur = undefined; break;
                        }
                        max --;
                    }
                    findingLink.push([cur, next[0]]);
                    if (cur != undefined && links[(i + len - 1) % len][0] == cur) { // && angle <= (polygon.length - 2) * Math.PI) {
                        let len = polygon.length;
                        let min = polygon[0];
                        let idx = 0;
                        for (let i = 0; i < len; i++) {
                            if (min > polygon[i]) {
                                idx = i;
                                min = polygon[i];
                            }
                        }
                        let temp = [];
                        for (let i = idx; i < idx + len; i++) {
                            temp.push(polygon[i % len]);
                        }
                        findedLink.push(...findingLink);
                        polygons.push(temp);
                        start.push(angleHurge);
                    }
                }
            }
        }

        let polygonsnew: IPolygon[] = [];
        polygons.forEach((polygon, idx) => {
            let len = polygon.length;
            let polygonnew: IPolygon = {
                triangles: [],
                points: [],
                min: [Number.MAX_SAFE_INTEGER, Number.MAX_SAFE_INTEGER],
                max: [Number.MIN_SAFE_INTEGER, Number.MIN_SAFE_INTEGER],
            }
            let idxstart = polygon.indexOf(start[idx]);
            for (let i = 0; i < len - 2; i++) {
                let a = polygon[idxstart];
                let b = polygon[(idxstart + i + 1) % len];
                let c = polygon[(idxstart + i + 2) % len];
                polygonnew.triangles.push([a, b, c]);
                let ap = data.get(a);
                let bp = data.get(b);
                let cp = data.get(c);
                polygonnew.min[0] = Math.min(ap[0], bp[0], cp[0], polygonnew.min[0]);
                polygonnew.min[1] = Math.min(ap[1], bp[1], cp[1], polygonnew.min[1]);
                polygonnew.max[0] = Math.max(ap[0], bp[0], cp[0], polygonnew.max[0]);
                polygonnew.max[1] = Math.max(ap[1], bp[1], cp[1], polygonnew.max[1]);
            }
            polygonnew.points = polygon;
            polygonsnew.push(polygonnew);
        });
        return polygonsnew;
    }
    protected findLinkNextClockwise(linking: TVerticeID, links: TLink[]): [TLink, number] | undefined {
        if (links) {
            let idx = links.findIndex((a) => a[0] == linking);
            let len = links.length;
            if (idx >= 0) {
                let link = links[(idx + len + 1) % len];
                // if (includePoints.indexOf(link[0]) >= 0) {
                //     return [link, link[3]];
                // } else {
                //     return undefined;
                // }
                return [link, link[3]];
                // for (let i = 0; i < len - 1; i++) {
                //     let link = links[(idx + i + len + 1) % len];
                //     if (includePoints.indexOf(link[0]) >= 0) {
                //         return [link, link[3]];
                //     }
                // }
            } else {
                return undefined;
            }
        }
        return undefined;
    }
    public static robbinBuild(resultPath: any, pointOffset: number, positions: Float32Array, colors: Float32Array, uvs: Float32Array, indices: Uint16Array) {
        let pointCount = resultPath.points.length;
        let idx = pointOffset;
        let vertexCountPerPoint = 3;
        let indiceIndexPerPoint = 6;
        let indiceCountPerPoint = 12;
        let maxDistance = resultPath.points[pointCount - 1].distance;
        for (let i = 0; i < pointCount; i++) {
            let point = resultPath.points[i];
            let x0 = point.position.x;
            let y0 = point.position.y;
            let z0 = point.position.z;
            let xs = point.tanget.x * point.size;
            let ys = point.tanget.y * point.size;
            let zs = point.tanget.z * point.size;
            let index = idx * vertexCountPerPoint;
            {
                positions[(index + 0) * 3 + 0] = x0 + xs;
                positions[(index + 0) * 3 + 1] = y0 + ys;
                positions[(index + 0) * 3 + 2] = z0 + zs;
                positions[(index + 1) * 3 + 0] = x0 - 0;
                positions[(index + 1) * 3 + 1] = y0 - 0;
                positions[(index + 1) * 3 + 2] = z0 - 0;
                positions[(index + 2) * 3 + 0] = x0 - xs;
                positions[(index + 2) * 3 + 1] = y0 - ys;
                positions[(index + 2) * 3 + 2] = z0 - zs;
            }
            if (colors){
                // let pp = Math.abs(point.distance / maxDistance - 0.5);
                // let edge = 0.48;
                // pp = 0.85 * (pp < edge ? 1.0 : pp / (0.5 - pp));
                let pp = 0.85;
                colors[(index + 0) * 4 + 0] = 1;
                colors[(index + 0) * 4 + 1] = 1;
                colors[(index + 0) * 4 + 2] = 1;
                colors[(index + 0) * 4 + 3] = 1;
                colors[(index + 1) * 4 + 0] = 1;
                colors[(index + 1) * 4 + 1] = 1;
                colors[(index + 1) * 4 + 2] = 1;
                colors[(index + 1) * 4 + 3] = pp;
                colors[(index + 2) * 4 + 0] = 0;
                colors[(index + 2) * 4 + 1] = 0;
                colors[(index + 2) * 4 + 2] = 0;
                colors[(index + 2) * 4 + 3] = 0;
            }
            if (uvs){
                uvs[(index + 0) * 2 + 0] = point.distance;
                uvs[(index + 0) * 2 + 1] = 0;
                uvs[(index + 1) * 2 + 0] = point.distance;
                uvs[(index + 1) * 2 + 1] = 0.5;
                uvs[(index + 2) * 2 + 0] = point.distance;
                uvs[(index + 2) * 2 + 1] = 1;
            }
            if (i > 0)
            {
                let preIndex = (idx - 1) * vertexCountPerPoint;
                let indiceIndex = (idx - 1) * indiceCountPerPoint;
                indices[indiceIndex + 0] = preIndex + 0;
                indices[indiceIndex + 1] = preIndex + 1;
                indices[indiceIndex + 2] = preIndex + 3;
    
                indices[indiceIndex + 3] = preIndex + 3;
                indices[indiceIndex + 4] = preIndex + 1;
                indices[indiceIndex + 5] = preIndex + 4;
                
                indices[indiceIndex + 6] = preIndex + 1;
                indices[indiceIndex + 7] = preIndex + 2;
                indices[indiceIndex + 8] = preIndex + 4;
                
                indices[indiceIndex + 9] = preIndex + 4;
                indices[indiceIndex +10] = preIndex + 2;
                indices[indiceIndex +11] = preIndex + 5;
            }
            idx += 1;
        }
    }
}

function buildGeometry(polygons: IPolygon[], data: IVertices, ignorePoints: TVerticeID[]): number[] {
    let position: number[] = [];
    polygons.forEach((polygon) => {
        let ignore = true;
        polygon.points.forEach((point) => {
            ignore = ignore && ignorePoints.indexOf(point) >= 0;
        });
        if (ignore) {
            polygon.triangles.forEach((triangle) => {
                triangle.forEach((point) => {
                    let vertex = data.get(point);
                    position.push(vertex[0], vertex[1], 0);
                });
            });
        }
    });
    return position;
}

//区域蒙版
export class AreaLock{
    private static points:IVertices;
    private static linkScene:CustomScene;
    public static polygons:IPolygon[];
    public static polygons2:IPolygon[];
    public static texsize: number = 256;
    public static maskScene:CustomScene;
    private static maskGeometry: CustomGeometry;
    private static quad: BABYLON.Mesh;
    private static links: Links;
    private static robbingeometry: CustomGeometry;
    private static basePath: BABYLON.Vector3[] = [];
    private static resultPath = { points: [] };
    private static areaMaterial: AreaMaterial;
    private static starMaterialList:any = [];//星空
    private static bgRepeatWidth: number;
    private static bgRepeatHeight: number;
    private static mapInfo: [number, number];//地图宽高
    //设置scene
    static setScene(scene:CustomScene){
        this.linkScene = scene;
        this.maskScene = new CustomRenderTargetScene("Polygon", this.texsize, this.texsize);
        this.maskScene.renderpipeline.opaqueRenderer.clearColor(0, 0, 0, 0);
        this.maskScene.active();
        // window.maskScene = this.maskScene;

        let mesh = new BABYLON.Mesh("Polygon", this.maskScene.impl);
        let geometry = new CustomGeometry(mesh, "Polygon", [Constants.VERTEX_ATTR_POSITION], 1000 * 3, Math.ceil(1000), false);
        this.maskGeometry = geometry;
        let material = new BABYLON.StandardMaterial("polygon", this.maskScene.impl);
        material.disableLighting = true;
        material.backFaceCulling = false;
        material.emissiveColor = new BABYLON.Color3(1, 1, 1);
        material.transparencyMode = 2;
        material.alphaMode = 2;
        material.alpha = 1;
        BABYLONAction.useMaterial(mesh, material, 2, 2);
        mesh.useVertexColors = false;
        mesh.hasVertexAlpha = false;
        mesh.alwaysSelectAsActiveMesh = true;
        mesh.isUnIndexed = true;

        let robbin = new BABYLON.Mesh("Polygonrobbin", this.maskScene.impl);
        BABYLONAction.useMaterial(robbin, material, 2, 2);
        this.robbingeometry = new CustomGeometry(robbin, "Polygonrobbin", [Constants.VERTEX_ATTR_POSITION, Constants.VERTEX_ATTR_COLOR4], 600, Math.ceil(600 / 3), true);
        
        robbin.useVertexColors = true;
        robbin.hasVertexAlpha = true;
        robbin.alwaysSelectAsActiveMesh = true;
        robbin.isUnIndexed = false;
        robbin.alphaIndex = 99999999;

        // window.robbin = robbin;
        
        let blurTarget = RenderTargetFactory.createRenderTarget(this.texsize, this.texsize, false);
        let irisblur = new DualBlur();
        this.maskScene.registerAfterRenderCall(() => {
            DualBlur.blend(RenderTargetFactory.getContext(),  0);
            irisblur.apply({ etype: KeyDualBlur, blurRadius: 1 / this.texsize, intensity: 1. }, [this.maskScene.renderTarget], blurTarget, RenderTargetFactory.getContext());
        });
        let _texture = new BABYLON.InternalTexture(this.linkScene.engine, 0);
        _texture.width = blurTarget.width;
        _texture.height = blurTarget.height;
        _texture.baseWidth = blurTarget.width;
        _texture.baseHeight = blurTarget.height;
        _texture.isReady = true;
        _texture._webGLTexture = blurTarget._webGLTexture;
        let texture = new BABYLON.Texture(null, this.linkScene.impl, true, true, BABYLON.Texture.LINEAR_LINEAR_MIPLINEAR);
        texture.delayLoadState = BABYLON.Constants.DELAYLOADSTATE_LOADED;
        texture._texture = _texture;
        (<any>texture)._delayedOnLoad = null;
        (<any>texture)._delayedOnError = null;

        texture.uScale = 1 ;
        texture.vScale = 1 ;

        let quad = BABYLON.PlaneBuilder.CreatePlane("PolygonMask", { size : 1 }, this.linkScene.impl);
        texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
        this.areaMaterial = new AreaMaterial("polygon", this.linkScene.impl);
        this.areaMaterial.maskTexture = texture;
        BABYLONAction.useMaterial(quad, this.areaMaterial, 2, 2);

        quad.alwaysSelectAsActiveMesh = true;
        quad.alphaIndex = 1000;
        quad.useVertexColors = true;
        quad.hasVertexAlpha = true;
        this.quad = quad;
        // window.quad = quad;
    };
    //初始化
    static init(pointPositionList:[number, number][], linePointList:[number, number][], mapInfo: [number, number, number, number]): CustomScene {
        this.resetPoints(pointPositionList, linePointList);
        {
            this.mapInfo = [mapInfo[2], mapInfo[3]];
            this.maskScene.resize(this.texsize, this.texsize);
            BABYLONAction.updateEngineDimensions(this.texsize, this.texsize);
            (<BABYLON.FreeCamera>this.maskScene.camera).piSizeForWidth = true;
            (<BABYLON.FreeCamera>this.maskScene.camera).piOrthSize = 1;
            (<BABYLON.FreeCamera>this.maskScene.camera).position.copyFromFloats(0, 0, -100);

            let geometry = this.maskGeometry;
            geometry.mesh.scaling.copyFromFloats(2 / mapInfo[2],  2 / mapInfo[3], 1);
            let [data] = geometry.vertices(Constants.VERTEX_ATTR_POSITION);
            let positions = buildGeometry(this.polygons, this.points, []);
            positions.forEach((val, idx) => {
                data[idx] = val;
            });
            geometry.update(positions.length / 3)
            
            this.robbingeometry.mesh.scaling.copyFromFloats(2 / mapInfo[2],  2 / mapInfo[3], 1);
            this.robbingeometry.update(0);

            // window.ssss = this.maskScene;

            this.quad.scaling.copyFromFloats(mapInfo[2], mapInfo[3], 1);
            // window.quad = this.quad;
        }
    }
    //设置背景
    static setRenderTexture(
        bgImage: string,
        gradBImage: string,
        bgRepeatWidth: number,
        bgRepeatHeight: number,
    ) {
        this.bgRepeatWidth = bgRepeatWidth;
        this.bgRepeatHeight = bgRepeatHeight;

        let bgtexture = new BABYLON.Texture(null, this.linkScene.impl, true, false, 2);
        let gradtexture = new BABYLON.Texture(null, this.linkScene.impl, true, false, 12);
        bgtexture.updateURL(bgImage);
        gradtexture.updateURL(gradBImage);
        gradtexture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
        gradtexture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
        this.areaMaterial.tilloff.copyFromFloats(bgRepeatWidth, bgRepeatHeight, 0, 0);
        this.areaMaterial.repeatBGTexture = bgtexture;
        this.areaMaterial.gradianBTexture = gradtexture;

        //星空
        this.setStarList(2);
    }
    //设置星空背景
    static setStarList(scale:number) {
        let starWidth = 2800 * scale;
        const [width, height] = this.mapInfo;
        
        let startX = -width/2;
        let startY = -height/2;
        // 生成坐标数组
        const coordinates:[number, number][] = [];
        while(startY <= height/2){
            startX = -width/2;
            while(startX <= width/2){
                coordinates.push([startX, startY]);
                startX += starWidth;
            }
            startY += starWidth;
        }
        coordinates.forEach((v, i)=>{
            let star = this.linkScene.prefabPool.createPrefab(`bahuangMapStarSky${i}`, "StarSky");
            star.alphaIndex = 1001;
            star.setEnabled(true);
            star.setScale(scale, scale, 1);
            star.setPosition(v[0], v[1], 1);
            this.starMaterialList.push(star);
        })
    }
    //设置背景scale
    static setBgScale(bgScaleRatio:number) {
        let minScale = 2;
        let targetScaleX = minScale * bgScaleRatio + (1 - bgScaleRatio) * this.bgRepeatWidth;
        let targetScaleY = minScale * bgScaleRatio + (1 - bgScaleRatio) * this.bgRepeatHeight;
        // console.log("==============>scale", targetScaleX, targetScaleY)
        if(targetScaleX === this.areaMaterial.tilloff.x){return}
        this.areaMaterial.tilloff.copyFromFloats(targetScaleX, targetScaleY, 0, 0);
    }
    //重新生成多边形数据
    static resetPoints(pointPositionList:[number, number][], linePointList:[number, number][]) {
        this.points = {
            data: pointPositionList,
            get: function(id: TVerticeID) {
                return this.data[id];
            }
        };
        this.links = new Links(linePointList , this.points);
        this.polygons = this.links.polygons(this.points);
        // this.polygons2 = this.links.polygons2(this.points);
        // console.error(this.polygons);
        // console.error(this.links);
    }
    //更新解锁点
    static updateUnlockPoints(ignores: TVerticeID[]) {
        let [data] = this.maskGeometry.vertices(Constants.VERTEX_ATTR_POSITION);
        let positions = buildGeometry(this.polygons, this.points, ignores);
        positions.forEach((val, idx) => {
            data[idx] = val;
        });
        this.maskGeometry.update(positions.length / 3);
        let edge = this.links.findEdge(this.points, ignores);
        if (edge.length == 1) {
            edge[0].points.forEach((idx) => {
                this.basePath.push(new BABYLON.Vector3(this.points.get(idx)[0], this.points.get(idx)[1], 0));
            });
            let idx = edge[0].points[0];
            this.basePath.push(new BABYLON.Vector3(this.points.get(idx)[0], this.points.get(idx)[1], 0));
            {
                let idx2 = edge[0].points[1];
                let x0 = this.points.get(idx)[0];
                let x1 = this.points.get(idx2)[0];
                let y0 = this.points.get(idx)[1];
                let y1 = this.points.get(idx2)[1];
                this.basePath.push(new BABYLON.Vector3(x0 * 0.9 + x1 * 0.1, y0 * 0.9 + y1 * 0.1, 0));
            }
            RibbonBuilder.FormatPathForceNormal(this.basePath, 100, this.resultPath, false, 0.1, new BABYLON.Vector3(0, 0, 1));
            let count = this.resultPath.points.length;
            let [data] = this.robbingeometry.vertices(Constants.VERTEX_ATTR_POSITION);
            let [color] = this.robbingeometry.vertices(Constants.VERTEX_ATTR_COLOR4);
            let indices = this.robbingeometry.indices();
            Links.robbinBuild(this.resultPath, 0, data, color, undefined, indices);
            this.robbingeometry.update(Math.max(count - 1, 0) * 12)
        }
    }

   
    //坐标点在那个多边形内
    static getPolygon(pos:[number, number]){
        let polygoncount = this.polygons.length;
        for (let i = 0; i < polygoncount; i++) {
            let polygon = this.polygons[i];
            if (isInPolygon(pos, polygon, this.points, isInMinMax)) {
                return polygon;
            }
        }
    }
    //销毁
    static dispose(){
        if (this.starMaterialList.length > 0) {
            this.starMaterialList.forEach(v=>v.dispose());
            this.starMaterialList.length = 0;
        }
    }
}


function isInMinMax(min: TPoint, max: TPoint, point: TPoint) {
    return (min[0] <= point[0] && min[1] <= point[1]) && (point[0] <= max[0] && point[1] <= max[1]);
}
function crossPoint(a: TPoint, b0: TPoint, b1: TPoint): number {
    return (b0[0] - a[0]) * (b1[1] - a[1]) - (b0[1] - a[1]) * (b1[0] - a[0]);
}
function isInTriangles(a: TPoint, b: TPoint, c: TPoint, p: TPoint): boolean {
    let aa = a;
    let bb = b;
    let cc = c;
    if (crossPoint(a, b, c) < 0) {
        cc = b;
        bb = c;
    }
    return crossPoint(aa, bb, p) > 0 && crossPoint(bb, cc, p) > 0 && crossPoint(cc, aa, p) > 0;
}

function isInPolygon(p: TPoint, polygon: IPolygon, data: IVertices, isInMinMax: (min: TPoint, max: TPoint, point: TPoint) => boolean): boolean {
    let result = false;
    if (isInMinMax(polygon.min, polygon.max, p)) {
        let len = polygon.triangles.length;
        for (let i = 0; i < len; i++) {
            let t = polygon.triangles[i];
            if (isInTriangles(data.get(t[0]), data.get(t[1]), data.get(t[2]), p)) {
                result = true;
                break;
            }
        }
    }
    return result;
}