import { Point } from "../common/point";
import { IEventData } from "./screen_con";

/**
 * ui 事件数据类型、和 IEventData 是一样的
 */
export interface IPointerEvent extends PointerEvent {
    /** 原生事件数据 */
    sourceEvent?: PointerEvent;
    /**
     * 事件是否从UI穿透
     */
    isSendNextLayer?: boolean
}

/**
 * 事件数据转换 支持 模拟数据、原生事件数据、gui直接传递的数据
 */
export type TTransEventData = IPointerEvent | IEventData;

/**
 * mesh 相同渲染组下，渲染顺序
 */
export enum EMeshRenderAlphaIndex {
    BuildGround = 3900,
    BuildPlane = 4000,
    BuildSpine = 5000,
    BuildModel = 5100,
}

/**
 * 建筑 scene渲染数据
 */
export interface IPointSceneRender {
    /**
     * id
     */
    pointId: string,
    /**
     * 位置
     */
    position: Point;
    /**
     * 缩放
     */
    scaling: Point;
    /**
     * 渲染cellId
     */
    renderCell: number;
    /**
     * 建筑渲染图名称
     */
    renderPoint: ERenderPointImageName;
    /**
     * 渲染模式
     */
    renderMode: EBuildRenderMode;
    /**
     * 最小渲染比例尺，大于才渲染
     */
    renderMinScale,
     /**
     * 最大渲染比例尺，小于才渲染
     */
    renderMaxScale
    /**
     * 是否显示加载
     */
    isEnabledLoad: boolean;
    /**
     * 渲染的图片
     */
    icon: string;
}

export enum ERenderPointImageName {
    None  = "",
    One   = "build_1"
}

/**
 * 可命中模型数据
 */
export interface IPickMesh<T> {
    _I_PICK: {
        /** 模型唯一id, 对应的uuid，可以为空 */
        pickId?: string;
        /** 事件类型 对应 MeshPickManage 中同名的静态方法 */
        pickEventKey?: string;
        /** 事件数据 */
        data?: T;
    }
}

/**
 * cellId 渲染参数
 */
export interface ICellIdOptions {
    cellHCount: number;
    cellVCount: number;
    /**
     * 纹理路径
     */
    texturePath: string;
}

/**
 * 基础渲染数据
 */
export interface IBaseRenderData {
    /**
     * gltf 所在文件夹
     */
    gltfPath: string;
    /**
     * gltf 名称
     */
    gltfName: string;
    /**
     * gltf 模型坐标
     */
    position: [number, number, number];
    /**
     * gltf 模型旋转
     */
    rotation?: [number, number, number];
    /**
     * gltf 模型缩放
     */
    scaling?: [number, number, number];
    /**
     * 公告板模式
     */
    billboardMode?: boolean;
    /**
     * 唯一id
     */
    uuid?: string;
    /**
     * 是否显示加载
     */
    isEnabledLoad?: boolean;
    /**
     * 是否禁用renderGroupId设置
     */
    disableRenderGroupId?: boolean;

    renderingGroupId?: number;

    alphaIndex?: number;
}

export interface IMineRenderData extends IBaseRenderData {
    /**
     * 最小渲染比例尺，大于才渲染
     */
    renderMinScale?: number;
    /**
     * 最大渲染比例尺，小于才渲染
     */
    renderMaxScale?: number;
    /**
     * 是否显示加载
     */
    isEnabledLoad?: boolean;
    /**
     * 是否固定缩放
     */
    isFixedScale?: boolean;
    /**
     * 事件响应的数据
     */
    pickEventData?: number;
}

/**
 * render 数据更新
 */
export interface IRenderUpdateData {
    /**
     * 位置数据
     */
    position?: [number, number, number];
    /**
     * 旋转数据
     */
    rotation?: [number, number, number];
    /**
     * 缩放数据
     */
    scaling?: [number, number, number];
    /**
     * 是否显示隐藏 
     */
    isEnabled?: boolean;
}

/**
 * 渲染动画数据
 */
export interface IRenderDataAnimation {
    /**
     * 唯一id
     */
    uuid: string;
    opt: {
        animName: string;
        isLoop: boolean;
        delay?: number;
        speed?: number;
        endCall?: (data: any) => void;
        startCall?: (data: any) => void;
        loopCall?: (data: any) => void;
        frameCall?: (framedata: any) => void;
    };
    /**
     * 是否有粒子
     */
    isParticle: boolean;
    /**
     * 是否自动播放回收的粒子
     */
    isAutoParticle?: boolean;
    /**
     * 下一个动画参数 start + idle 时用
     */
    nextAnimaOpt?: IRenderDataAnimation;
    /**
     * 回调函数
     */
    callBack?: () => void;
}

/**
 * 渲染物加载时，需要传入的事件数据
 */
export interface IRenderEventData<T> {
    /** 模型唯一id, 对应的uuid，可以为空 */
    pickId?: string;
    /** 事件类型 对应 MeshPickManage 中同名的静态方法 */
    pickEventKey?: string;
    /** 事件自定义数据 */
    data?: T;
    /** 事件上指定过滤掉一些节点 */
    filter?: string[];
    /** 事件重要级别（临时的，如果有这个属性，就将当前mesh置顶） */
    important?: number;
}

/**
 * 渲染加载参数
 */
export interface IRenderAddOptions<T> {
    /** 加载任务数据 */
    task: IBaseRenderData;
    /** 加载动画数据 */
    animaOpt: IRenderDataAnimation;
    /** 加载事件数据 */
    eventData?: IRenderEventData<T>;
}

export enum EFocusRange {
    NONE = 0,
    MAX = 9999,
    MINE_SCALE = 7,   // 推荐矿场视野

    MINE_MAX = 50,  // 最大矿场视野
    MINE_MAX_M = 15,  // 最大矿场视野
    MINE_MIN = 4,
}

export enum ERenderDataParam {
    /** 最大渲染值 */
    MaxScale = "renderMaxScale",
    /** 最小渲染值 */
    MinScale = "renderMinScale",
    /** 点击事件数据 */
    PickEventData = "pickEventData",
    /** 是否有设置的alphaIndex */
    AlphaIndex = "alphaIndex",
    /** 是否隐式加载 */
    EnabledLoad = "isEnabledLoad",
    /** 是否使用传入的renderGroupId设置 */
    RenderingGroupId = "renderingGroupId",
    /** 是否禁用renderGroupId设置 */
    DisableRenderGroupId = "disableRenderGroupId",
}

/**
 * 建筑渲染的不同模式
 */
export enum EBuildRenderMode {
    /** 无 */
    NONE = 0,
    /** 自定义 BuildRender 模式 */
    BuildRender = 1,
    /** spine model 模式 */
    SpineRender,
    /** modelObj 模式 */
    ModelRender,
    /** modelObj + buildRender */
    DoubleRender,
}

/**
 * 命中事件步骤
 */
export enum EPickStep {
    Down   = "down",
    Move   = "move",
    Up     = "up",
    Click  = "click",
    Cancel = "cancel",
    Check  = "check",
}

export enum ERenderDateName {
    model = 'model',
    panel = 'panel'
}