import { Point } from "../common/point";

/**
 * 屏幕数据
 */
export interface IScreenInfo {
    /**
     * 逻辑宽度
     */
    sourceWidth: number;
    /**
     * 逻辑高度
     */
    sourceHeight: number;
    /**
     * 布局宽度 = 逻辑宽度 * pixelRatio
     */
    layoutWidth: number;
    /**
     * 布局高度 = 逻辑高度 * pixelRatio
     */
    layoutHeight: number;
    /**
     * 屏幕像素比，通常在事件中，如果需要计算 pageX 之类的东西，需要乘上去
     */
    pixelRatio: number;
}

/**
 * 自定义事件，添加坐标转换成真实坐标
 */
export interface IEventData {
    /**
     * gui 坐标
     */
    x: number;
    /**
     * gui 坐标
     */
    y: number;
    /**
     * gui 坐标
     */
    clientX?: number;
    /**
     * gui 坐标
     */
    clientY?: number;
    /**
     * 缩放值
     */
    deltaX?: number;
    /**
     * x * pixelRatio
     */
    prx?: number;
    /**
     * y * pixelRatio
     */
    pry?: number;
    /**
     * 页面坐标 x
     */
    pageX?: number;
    /**
    * 页面坐标 y
    */
    pageY?: number;
    /**
     * 按键类型
     */
    button?: number;
    /**
     * 原生事件数据
     */
    sourceEvent?: PointerEvent;
    /**
     * 事件是否从UI穿透
     */
    isSendNextLayer?: boolean;
    /**
     * 多点事件数据堆 - 坐标堆
     * * key: 事件ID
     * * value: { x, y } 事件坐标
     */
    pointers?: { x: number; y: number; id: number, sourceEvent?: { clientX: number, clientY: number } }[];
}

/**
 * 鼠标各种事件回调
 */
export interface IMouseCallbacks {
    /**
     * 松开
     */
    onUp?: (event: IEventData) => void;
    /**
     * 按下
     */
    onDown?: (event: IEventData) => void;
    /**
     * 点击
     * @param quad 当前命中的 quad
     */
    onClick?: (event: IEventData) => void;
    /**
     * 滑动
     */
    onHover?: (event: IEventData) => void;
    /**
     * 鼠标离开
     */
    onLeave?: () => void;
    /**
     * 滚轮
     * @param targetScale 目标缩放值
     */
    onWheel?: (direction: "up" | "down", targetScale: number) => void;
    /**
     * 按下滑动
     * @param movex x 方向移动量
     * @param movey y 方向移动量
     */
    onDownMove?: (event: IEventData, movex: number, movey: number) => void;
    /**
     * 双指事件
     */
    onMultiPointer?: (sceneDistance: number, uiDistance: number) => void;
    /**
     * 场景点击
     */
    onSceneClick?: (event: IScene3DEventInfo, point: Point) => void;
    /**
     * 场景按下滑动
     */
    onSceneDownMove?: (event: IScene3DEventInfo, point: Point, movex: number, movey: number) => void;
    /**
     * 场景鼠标按下
     */
    onSceneMouseDown?: (event: IScene3DEventInfo, point: Point) => void;
    /**
     * 场景鼠标按下
     */
    onSceneMouseUp?: (event: IScene3DEventInfo, point: Point) => void;
}

/**
 * 自定义的场景命中数据，只是为了解决后端引用问题。没其他作用
 */
export interface IScene3DEventInfo {
    rayInfo: { pickedMesh?: any };
    sceneEvent: { x?: number, y?: number, clientX?: number, clientY?: number };
    uiEvent: IEventData;
    isSendNextLayer: boolean;
}